package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kbox2d.api.BodyDef
import ru.dbotthepony.kbox2d.api.FixtureDef
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.util.TwoDimensionalArray
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kstarbound.world.phys.RectTileFlooderDepthFirst
import ru.dbotthepony.kstarbound.world.phys.RectTileFlooderSizeFirst
import ru.dbotthepony.kvector.arrays.Object2DArray
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.Vector2d
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashSet

/**
 * Чанк мира
 *
 * Хранит в себе тайлы и ентити внутри себя
 *
 * Считается, что один тайл имеет форму квадрата и сторона квадрата примерно равна полуметру,
 * что будет называться Starbound Unit
 *
 * Весь игровой мир будет измеряться в Starbound Unit'ах
 */
abstract class Chunk<WorldType : World<WorldType, This>, This : Chunk<WorldType, This>>(val world: WorldType, val pos: ChunkPos) {
	/**
	 * Возвращает общий счётчик изменений чанка
	 */
	var changeset = 0
		private set

	/**
	 * Возвращает счётчик изменений чанка по тайлам
	 */
	var tileChangeset = 0
		private set

	/**
	 * Возвращает счётчик изменений чанка по жидкостям
	 */
	var liquidChangeset = 0
		private set

	val left get() = pos.left
	val right get() = pos.right
	val top get() = pos.top
	val bottom get() = pos.bottom
	val topLeft get() = pos.topLeft
	val topRight get() = pos.topRight
	val bottomLeft get() = pos.bottomLeft
	val bottomRight get() = pos.bottomRight

	val aabb = aabbBase + Vector2d(pos.x * CHUNK_SIZE.toDouble(), pos.y * CHUNK_SIZE.toDouble())

	var isPhysicsDirty = false

	fun markPhysicsDirty() {
		if (isPhysicsDirty)
			return

		isPhysicsDirty = true
		world.dirtyPhysicsChunks.add(this as This)
	}

	fun bakeCollisions() {
		foreground.bakeCollisions()
	}

	inner class LiquidState(val def: LiquidDefinition) {
		var pressure: Float = 0f
			set(value) {
				if (value != field) {
					field = value
					liquidChangeset++
					changeset++
				}
			}

		var level: Float = 1f
			set(value) {
				if (value != field) {
					field = value
					liquidChangeset++
					changeset++
				}
			}

		var isInfinite: Boolean = false
			set(value) {
				if (value != field) {
					field = value
					liquidChangeset++
					changeset++
				}
			}
	}

	inner class TileLayer : IMutableTileChunk {
		/**
		 * Возвращает счётчик изменений этого слоя
		 */
		var changeset = 0
			private set

		private val collisionCache = ArrayList<AABB>()
		private val collisionCacheView = Collections.unmodifiableCollection(collisionCache)
		private var collisionChangeset = -1

		private val body = world.physics.createBody(BodyDef(
			position = pos.firstBlock.toDoubleVector(),
			userData = this
		))

		private val collisionChains = ArrayList<B2Fixture>()

		fun bakeCollisions() {
			if (collisionChangeset == changeset)
				return

			collisionChangeset = changeset
			collisionCache.clear()

			for (box in collisionChains) {
				body.destroyFixture(box)
			}

			collisionChains.clear()

			val xyAdd = Vector2d(pos.x * CHUNK_SIZEd, pos.y * CHUNK_SIZEd)
			val seen = BooleanArray(CHUNK_SIZE * CHUNK_SIZE)

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0..CHUNK_SIZE_FF) {
					if (!seen[x or (y shl CHUNK_SHIFT)] && tiles[x, y].material != null) {
						val depthFirst = RectTileFlooderDepthFirst(tiles, seen, x, y)
						val sizeFirst = RectTileFlooderSizeFirst(tiles, seen, x, y)
						val xSpanDepth = depthFirst.maxs.x - depthFirst.mins.x
						val ySpanDepth = depthFirst.maxs.y - depthFirst.mins.y
						val xSpanSize = sizeFirst.maxs.x - sizeFirst.mins.x
						val ySpanSize = sizeFirst.maxs.y - sizeFirst.mins.y

						val aabb: AABB

						if (xSpanDepth * ySpanDepth > xSpanSize * ySpanSize) {
							depthFirst.markSeen()
							aabb = AABB(depthFirst.mins.toDoubleVector(), depthFirst.maxs.toDoubleVector() + Vector2d.POSITIVE_XY)
						} else {
							sizeFirst.markSeen()
							aabb = AABB(sizeFirst.mins.toDoubleVector(), sizeFirst.maxs.toDoubleVector() + Vector2d.POSITIVE_XY)
						}

						collisionChains.add(body.createFixture(FixtureDef(
							shape = PolygonShape().also { it.setAsBox(aabb.width / 2.0, aabb.height / 2.0, aabb.centre, 0.0) },
							friction = 0.4,
						)))

						collisionCache.add(aabb + xyAdd)
					}
				}
			}
		}

		/**
		 * Возвращает список AABB тайлов этого слоя
		 *
		 * Данный список напрямую указывает на внутреннее состояние и будет изменён при перестройке
		 * коллизии чанка, поэтому если необходим стабильный список, его необходимо скопировать
		 */
		fun collisionLayers(): Collection<AABB> {
			if (collisionChangeset != changeset) {
				bakeCollisions()
			}

			return collisionCacheView
		}

		override val pos: ChunkPos
			get() = this@Chunk.pos

		inline val left get() = pos.left
		inline val right get() = pos.right
		inline val top get() = pos.top
		inline val bottom get() = pos.bottom
		inline val topLeft get() = pos.topLeft
		inline val topRight get() = pos.topRight
		inline val bottomLeft get() = pos.bottomLeft
		inline val bottomRight get() = pos.bottomRight

		/**
		 * Хранит тайлы как x + y * CHUNK_SIZE
		 */
		private val tiles: Object2DArray<TileState> = Object2DArray(CHUNK_SIZE, CHUNK_SIZE, TileState.EMPTY)

		override fun getTile(x: Int, y: Int): TileState {
			return tiles[x, y]
		}

		override fun setTile(x: Int, y: Int, value: TileState) {
			tiles[x, y] = value
			tileChangeset++
			changeset++
			collisionChangeset++
			this@Chunk.changeset++
		}

		override fun randomLongFor(x: Int, y: Int): Long {
			return super.randomLongFor(x, y) xor world.seed
		}
	}

	val foreground = TileLayer()
	val background = TileLayer()

	protected val liquidStates: TwoDimensionalArray<LiquidState> = TwoDimensionalArray(CHUNK_SIZE, CHUNK_SIZE)

	fun getLiquid(x: Int, y: Int) = liquidStates[x, y]

	fun setLiquid(x: Int, y: Int, value: LiquidDefinition?): LiquidState? {
		if (value == null) {
			val old = liquidStates.set(x, y, null)

			if (old != null) {
				changeset++
				liquidChangeset++
			}

			return old
		}

		val state = LiquidState(value)
		liquidStates[x, y] = state
		changeset++
		liquidChangeset++
		return state
	}

	protected val entities = HashSet<Entity>()
	val entitiesAccess: Set<Entity> = Collections.unmodifiableSet(entities)

	protected abstract fun onEntityAdded(entity: Entity)
	protected abstract fun onEntityTransferedToThis(entity: Entity, otherChunk: This)
	protected abstract fun onEntityTransferedFromThis(entity: Entity, otherChunk: This)
	protected abstract fun onEntityRemoved(entity: Entity)

	fun addEntity(entity: Entity) {
		if (!entities.add(entity)) {
			throw IllegalArgumentException("Already having having entity $entity")
		}

		onEntityAdded(entity)
	}

	fun transferEntity(entity: Entity, otherChunk: Chunk<*, *>) {
		if (otherChunk == this)
			throw IllegalArgumentException("what?")

		if (this::class.java != otherChunk::class.java) {
			throw IllegalArgumentException("Incompatible types: $this !is $otherChunk")
		}

		if (!entities.add(entity)) {
			throw IllegalArgumentException("Already containing $entity")
		}

		onEntityTransferedToThis(entity, otherChunk as This)
		otherChunk.onEntityTransferedFromThis(entity, this as This)

		if (!otherChunk.entities.remove(entity)) {
			throw IllegalStateException("Unable to remove $entity from $otherChunk after transfer")
		}
	}

	fun removeEntity(entity: Entity) {
		if (!entities.remove(entity)) {
			throw IllegalArgumentException("Already not having entity $entity")
		}

		onEntityRemoved(entity)
	}

	override fun toString(): String {
		return "Chunk(pos=$pos, entityCount=${entities.size}, world=$world)"
	}

	companion object {
		private val aabbBase = AABB(
			Vector2d.ZERO,
			Vector2d(CHUNK_SIZE.toDouble(), CHUNK_SIZE.toDouble()),
		)
	}
}
